import enum
import json
import os
import pathlib
import shutil
import socket
import sys
import time
from typing import Any, Optional, Sequence

import np_config
import np_logging
import requests

import np_services.utils as utils
from np_services.protocols import TestError

# global vars -------------------------------------------------------------------------- #
logger = np_logging.getLogger(__name__)  # logs will show full module path

__name = "OpenEphys"  # Service protocol operations will see just the 'class' name

exc: Optional[BaseException] = None
initialized: float = 0
"`time.time()` when the service was initialized."

try:
    host: str = np_config.Rig().Acq
except ValueError:
    logger.warning("Not connected to a rig: `OpenEphys.host` needs to be set manually.")
port: str | int = 37497  # 1-800-EPHYS
latest_start: float = 0
"`time.time()` when the service was last started via `start()`."

# for launching:
rsc_app_id = "open-ephys"

# device records:
gb_per_hr: int | float = 250  # per drive
min_rec_hr: int | float = 2
pretest_duration_sec: int | float = 0.5

# for resulting data:
folder: str  #! required
"The string that will be sent to Open Ephys to name the recording: typically `0123456789_366122_20220618`"
data_files: list[pathlib.Path] = []
"Storage for paths collected over the experiment."
data_root: Optional[pathlib.Path] = None
# -------------------------------------------------------------------------------------- #


class State(enum.Enum):
    idle = "IDLE"
    acquire = "ACQUIRE"
    record = "RECORD"


class Endpoint(enum.Enum):
    status = "status"
    recording = "recording"
    processors = "processors"
    message = "message"


def launch() -> None:
    utils.start_rsc_app(host, rsc_app_id)


def pretest() -> None:
    logger.info("OpenEphys | Starting pretest")
    global folder
    folder = "_pretest_"
    initialize()
    test()
    try:
        start()
        time.sleep(pretest_duration_sec)
        verify()
    finally:
        stop()
    finalize()
    validate()
    logger.info("OpenEphys | Pretest passed")

def url(endpoint: Endpoint):
    return f"http://{host}:{port}/api/{endpoint.value}"


def get_state() -> requests.Response:
    mode = requests.get(u := url(Endpoint.status)).json().get("mode")
    logger.debug("%s -> get mode: %s", u, mode)
    return mode


def set_state(state: State) -> requests.Response:
    msg = {"mode": state.value}
    mode = requests.put(u := url(Endpoint.status), json.dumps(msg))
    logger.debug("%s <- set mode: %s", u, state.value)
    return mode


def is_connected() -> bool:
    global exc

    if not utils.is_online(host):
        exc = TestError(
            f"OpenEphys | No response from {host}: may be offline or unreachable"
        )
        return False

    try:
        state = get_state()
    except requests.RequestException:
        exc = TestError(
            f"OpenEphys | No response from Open Ephys http server: is the software started?"
        )
        return False
    else:
        if not any(_.value == state for _ in State):
            exc = TestError(f"OpenEphys | Unexpected state: {state}")
            return False

    return True


def initialize() -> None:
    logger.info("OpenEphys | Initializing")
    global data_files
    data_files = []
    global initialized
    initialized = time.time()
    global folder
    set_folder(folder)


def get_required_disk_gb() -> int | float:
    """The minimum amount of free disk space required to start recording."""
    return gb_per_hr * min_rec_hr


def is_disk_space_ok() -> bool:
    global exc
    exc = None
    required = get_required_disk_gb()
    for data_root in get_data_roots():
        try:
            free = utils.free_gb(data_root)
        except FileNotFoundError as e:
            exc = e
            logger.exception(f"{__name} data path not accessible: {data_root}")
        else:
            logger.info(
                "%s free disk space on %s: %s GB", __name, data_root.drive, free
            )
            if free < required:
                exc = ValueError(
                    f"{__name} free disk space on {data_root.drive} doesn't meet minimum of {required} GB"
                )
    if exc:
        return False
    return True


def test() -> None:
    logger.info("OpenEphys | Testing")
    if not is_connected():
        if exc:
            raise exc
    gb = get_required_disk_gb()
    if not is_disk_space_ok():
        raise TestError(
            f"{__name} free disk space on one or more recording drives doesn't meet minimum of {gb} GB"
        ) from exc
    unlock_previous_recording()


def is_started() -> bool:
    if get_state() == State.record.value:
        return True
    return False


def is_ready_to_start() -> bool:
    if get_state() == State.acquire.value:
        return True
    return False


def start() -> None:
    logger.info("OpenEphys | Starting recording")
    if is_started():
        logger.warning("OpenEphys is already started")
        return
    if not is_ready_to_start():
        set_state(State.acquire)
        time.sleep(0.5)
    global latest_start
    latest_start = time.time()
    set_state(State.record)


def stop() -> None:
    logger.info("OpenEphys | Stopping recording")
    set_state(State.acquire)


def finalize() -> None:
    logger.info("OpenEphys | Finalizing")
    data_files.extend(get_latest_data_dirs())
    unlock_previous_recording()


def set_folder(
    name: str, prepend_text: str = "", append_text: str = ""
) -> None:
    """Recording folder string"""
    recording = requests.get(url(Endpoint.recording)).json()

    if name == "":
        name = "_"
        logger.warning(
            "OpenEphys | Recording directory cannot be empty, replaced with underscore: %s",
            name,
        )
    if "." in name:
        name.replace(".", "_")
        logger.warning(
            "OpenEphys | Recording directory cannot contain periods, replaced with underscores: %s",
            name,
        )

    recording["base_text"] = name
    recording["prepend_text"] = prepend_text
    recording["append_text"] = append_text
    logger.debug(
        "OpenEphys | Setting recording directory to: %s",
        prepend_text + name + append_text,
    )
    response = requests.put(url(Endpoint.recording), json.dumps(recording))
    time.sleep(0.1)
    if (actual := response.json().get("base_text")) != name:
        raise TestError(
            f"OpenEphys | Set folder to {name}, but software shows: {actual}"
        )


def get_folder() -> str | None:
    return requests.get(url(Endpoint.recording)).json().get("base_text")


def clear_open_ephys_name() -> None:
    set_folder("_temp_")


def set_idle():
    "Should be called before sending any configuration to Open Ephys"
    if is_started():
        stop()
    time.sleep(0.5)
    set_state(State.idle)


def unlock_previous_recording():
    "stop rec/acquiring | set name to _temp_ | record briefly | acquire | set name to folder"
    logger.debug("OpenEphys | Unlocking previous recording")
    set_idle()
    time.sleep(0.5)
    clear_open_ephys_name()
    time.sleep(0.5)
    start()
    time.sleep(0.5)
    stop()
    time.sleep(0.5)
    global folder
    set_folder(folder)


def get_record_nodes() -> list[dict[str, Any]]:
    """Returns a list of record node info dicts, incl keys `node_id`, `parent_directory`"""
    return requests.get(url(Endpoint.recording)).json().get("record_nodes", None) or []


def get_data_roots() -> list[pathlib.Path]:
    return [
        pathlib.Path(f"//{host}/{_['parent_directory'].replace(':','')}")
        for _ in get_record_nodes()
    ]


def get_latest_data_dirs() -> list[pathlib.Path]:
    """Returns the path to the latest data folder, based on the latest modified time"""
    dirs = []
    for root in get_data_roots():
        if subfolders := [
            sub
            for sub in root.iterdir()
            if sub.is_dir()
            and not any(
                _ in str(sub) for _ in ["System Volume Information", "$RECYCLE.BIN"]
            )
        ]:
            subfolders.sort(key=lambda _: _.stat().st_ctime)
            dirs.append(subfolders[-1])
    return dirs

def check_files_increasing_in_size() -> None:
    for data_dir in get_latest_data_dirs():
        for file in reversed(
            utils.get_files_created_between(
                data_dir, "*/*/*/continuous/*/sample_numbers.npy", latest_start
            )
        ):
            if utils.is_file_growing(file):
                break
        else:
            raise TestError(
                f"OpenEphys | Data file(s) not increasing in size in {data_dir}"
            )

def verify() -> None:
    logger.debug("OpenEphys | Verifying")
    check_files_increasing_in_size()    
    logger.info(
        "OpenEphys | Verified files are increasing in size for all Record Nodes"
    )


def validate() -> None:
    logger.warning("OpenEphys | validate() not implemented")


def set_ref(ext_tip="TIP"):
    # for port in [0, 1, 2]:
    #     for slot in [0, 1, 2]:

    slot = 2  #! Test
    port = 1  #! Test
    dock = 1  # TODO may be 1 or 2 with firmware upgrade
    tip_ref_msg = {"text": f"NP REFERENCE {slot} {port} {dock} {ext_tip}"}
    # logger.info(f"sending ...
    # return
    requests.put(
        "http://localhost:37497/api/processors/100/config", json.dumps(tip_ref_msg)
    )
    time.sleep(3)


# TODO set up everything possible from here to avoid accidentally changed settings ?
# probe channels
# sampling rate
# tip ref
# signal chain?
# acq drive letters

"""
if __name__ == "__main__":
    r = requests.get(Endpoint.recording)
    print((r.json()['current_directory_name'], r.json()['prepend_text'], r.json()['append_text']))
        
    r = EphysHTTP.set_folder(path = "mouseID_", prepend_text="sessionID", append_text="_date")
    print((r.json()['current_directory_name'], r.json()['prepend_text'], r.json()['append_text']))
    
    r = EphysHTTP.set_folder(path = "mouse", prepend_text="session", append_text="date")
    print((r.json()['current_directory_name'], r.json()['prepend_text'], r.json()['append_text']))
    
    print((r.json()['base_text'])) # fails as of 06/23 https://github.com/open-ephys/plugin-GUI/pull/514
"""



import argparse
import json
import logging
import os
import sys
from pprint import pformat, pprint
from socket import *

import np_logging

logger = np_logging.getLogger(__name__)

R = {"mvr_request": ""}
encoding = "utf-8"


class ResponseBuffer:
    def __init__(self):
        self.read_buffer = []

    def parse_buffer(self, buf):
        if not buf:
            return []
        buf = buf.decode()
        read_bracket_count = 0
        self.read_buffer.extend(buf)
        count = 0
        messages = []
        for i, c in enumerate(
            self.read_buffer
        ):  # should maintain an internal pointer so it doesn't redo the list
            count += 1
            # update the "bracket stack"
            if c == "{":
                read_bracket_count += 1
            elif c == "}":
                read_bracket_count -= 1

            if read_bracket_count == 0:  # a full JSON string is available
                try:
                    messages.append(
                        json.loads("".join(self.read_buffer[i - count + 1 : i + 1]))
                    )
                except TypeError:
                    logger.warning(
                        "%s | Error parsing message: %s",
                        __class__.__name__,
                        self.read_buffer,
                    )
                count = 0

        # strip prior json messages off the buffer
        if count == 0 and read_bracket_count == 0:  # There must be a better way
            self.read_buffer = []
        else:
            self.read_buffer = self.read_buffer[-count:]

        return messages


class MVRConnector:
    def __init__(self, args=None):
        self.response_buffer = ResponseBuffer()
        self.device_index_map = {}
        self._errors_since_last_success: int = 0
        self._delete_on_copy: bool = True
        self._recording: bool = False
        self._args = args
        self._mvr_sock = None
        self._host_to_camera_map = {}
        self._mvr_connected = False
        self.comp_ids = []  # temporary to allow for some debugging
        self.output_dir = "c$/ProgramData/AIBS_MPE/MVR/data/"
        try:
            self.connect_to_mvr()
        except Exception as err:
            logger.error(f"failed to connect to mvr:{err}")
            exit()

    def _recv(self):
        if not self._mvr_connected:
            self.connect_to_mvr()
            return

        try:
            ret_val = self._mvr_sock.recv(1024)
        except ConnectionResetError as e:
            logger.warning("%s | Connection reset error", __class__.__name__)
            self._mvr_connected = False
            return []
        except:
            return []
        return ret_val

    def read(self):
        buf = self._recv()
        return self.response_buffer.parse_buffer(buf)

    def _send(self, msg):
        """
        msg is a dictionary.
        _send creates json from the dictionary and sends it as a byte object
        """
        msg = json.dumps(msg).encode()
        logger.debug("%s | Sending: %s", __class__.__name__, msg)
        if not self._mvr_connected:
            self.connect_to_mvr()
        if not self._mvr_connected:
            return
        try:
            self._mvr_sock.send(msg)
        except (ConnectionResetError, ConnectionRefusedError):
            self._mvr_connected = False

    def connect_to_mvr(self):
        """
        Creates a STREAM Socket connection to the MultiVideoRecorder
        """
        self._mvr_sock = socket(AF_INET, SOCK_STREAM)
        self._mvr_sock.settimeout(10.0)
        if self._errors_since_last_success == 0:
            logger.debug(
                "%s | Connecting on %s:%s",
                __class__.__name__,
                self._args["host"],
                self._args["port"],
            )
        try:
            self._mvr_sock.connect((self._args["host"], self._args["port"]))
            self._mvr_connected = True
        except OSError:
            if self._errors_since_last_success == 0:
                logger.debug(
                    "%s | Connection failed, will be re-attempted on the next write.",
                    __class__.__name__,
                )
            self._errors_since_last_success += 1
        else:
            self._errors_since_last_success = 0
            logger.debug("%s | Connection success: %s", __class__.__name__, self.read())

    def get_version(self):
        msg = {"mvr_request": "get_version"}
        self._send(msg)
        return self.read()

    def start_display(self):
        msg = {"mvr_request": "start_display"}
        self._send(msg)
        return self.read()

    def stop_display(self):
        msg = {"mvr_request": "stop_display"}
        self._send(msg)
        return self.read()

    def start_record(self, file_name_prefix="", sub_folder=".", record_time=4800):
        self._send(
            {
                "mvr_request": "start_record",
                "sub_folder": sub_folder,
                "file_name_prefix": file_name_prefix,
                "recording_time": record_time,
            }
        )

    def start_single_record(
        self, host, file_name_prefix="", sub_folder=".", record_time=4800
    ):
        print(f"start single record on {host}")
        if host not in self._host_to_camera_map:
            comp = self.host_to_comp["host"]
            logger.warning(
                f"Start Single Record: Can not find host {host} ({comp}) associated with a camera."
            )
            return

        self._send(
            {
                "mvr_request": "start_record",
                "camera_indices": [
                    {"camera_index": f"Camera {self._host_to_camera_map[host]}"}
                ],
                "sub_folder": sub_folder,
                "file_name_prefix": file_name_prefix,
                "recording_time": record_time,
            }
        )

    def set_automated_ui(self, state):
        if state:
            message = {"mvr_request": "set_automated_ui"}

        else:
            message = {"mvr_request": "set_unautomated_ui"}

        self._send(message)
        return self.read()

    def stop_record(self):
        msg = {"mvr_request": "stop_record"}
        self._send(msg)
        # return self.read()

    def stop_single_record(self, host):
        if host not in self._host_to_camera_map:
            comp = self.host_to_comp["host"]
            logger.warning(
                f"Stop Single Record: Can not find host {host} ({comp}) associated with a camera."
            )
            return

        cam_id = self._host_to_camera_map[host]
        message = {
            "mvr_request": "stop_record",
            "camera_indices": [f"Camera {cam_id}",],
        }
        self._send(message)

    def take_snapshot(self):
        self._send({"mvr_request": "take_snapshot"})

    def define_hosts(self, hosts):
        self._host_to_camera_map = {}
        for idx, host in enumerate(hosts):
            self._host_to_camera_map[host] = idx + 1

        # print(f'Host To Camera Map: {pformat(self._host_to_camera_map)}')
        self.host_to_comp = list(zip(hosts, self.comp_ids))

    def request_camera_ids(self):
        self._send({"mvr_request": "get_camera_ids"})
        return self.read()

    def highlight_camera(self, device_name):
        print(self.device_index_map)
        index = self.device_index_map[device_name]
        msg = {"mvr_request": "toggle_highlight_camera", "camera": index}
        self._send(msg)

    def unhighlight_camera(self, panel):
        msg = {"mvr_request": "unhighlight_panel"}
        self._send(msg)
        return self.read()

    def get_state(self):
        if self._recording:
            return "BUSY", "RECORDING"
        else:
            return "READY", ""

    def shutdown(self):
        self._rep_sock.close()

    def _onclose(self):
        self.shutdown()

    @property
    def platform_info(self):
        return "0.1.0"



import pathlib

import h5py
import pandas as pd
import pydantic
import tables

f = pathlib.Path('//allen/programs/mindscope/workgroups/np-exp/habituation/1253722691_661730_20230313/1253722691_661730_20230313.sync')

h = h5py.File(f, 'r')
t = tables.open_file(f, 'r')
pd.read_hdf(f.as_posix(), 'meta')




class Sync(pydantic.BaseModel):
    pass
    


import np_services.open_ephys as OpenEphys
from np_services.protocols import *
try:
    from np_services.proxies import *
except ValueError as exc:
    print(f"Error importing np_services.proxies: {exc}")
from np_services.utils import *



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['np_services']

package_data = \
{'': ['*']}

install_requires = \
['backoff',
 'fabric>=2.7,<3.0',
 'h5py>=3.8.0,<4.0.0',
 'np_config>=0.4.17',
 'np_session>=0.4.13',
 'pandas>=1.5.3,<2.0.0',
 'pydantic>=1.10,<2.0',
 'pyzmq',
 'requests>=2,<3',
 'tables>=3.8.0,<4.0.0']

setup_kwargs = {
    'name': 'np-services',
    'version': '0.1.47',
    'description': 'Tools for interfacing with devices and services used in Mindscope Neuropixels experiments at the Allen Institute.',
    'long_description': '# service usage\n![Services](./services.drawio.svg)',
    'author': 'bjhardcastle',
    'author_email': 'ben.hardcastle@alleninstitute.org',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.11,<4.0',
}


setup(**setup_kwargs)
